/*
 * anttrafficmonitor.{cc,hh} -- monitor network traffic - used by ForwardAnt
 * Ingebrigt Fuglem
 *
*/

#include <click/config.h>
#include <click/error.hh>
#include <click/confparse.hh>
#include "anttrafficmonitor.hh"
#include <click/ipaddress.hh>
#include <click/ewma.hh>
#include "anttable.hh"
#include <linux/netdevice.h> 
CLICK_DECLS

AntTrafficMonitor::AntTrafficMonitor() : Element(1, 1), _antTable(0), _x(0), _delta(0)
{
  _now = click_jiffies();
}

AntTrafficMonitor::~AntTrafficMonitor()
{
}

int
AntTrafficMonitor::configure(Vector<String> &conf, ErrorHandler *errh)
{
  int beta = 0;
  int res = cp_va_parse(conf, this, errh,
      cpElement, "AntTable element", &_antTable,
      cpUnsigned, "Interface no", &_if_no,
      cpReal10, "beta", 2, &beta,
      cpEnd);
  _beta = ((double)beta)/100;
  return res;
}

int
AntTrafficMonitor::initialize(ErrorHandler *errh)
{
  if(_antTable == 0){
    return errh->error("no AntTable argument");
  }
  return 0;
}


Packet *
AntTrafficMonitor::simple_action(Packet *p)
{
  assert(_antTable);
  
  // Measure byte and packet rate
  _packet_rate.update(1);
  _byte_rate.update(p->length());
  
  unsigned now = click_jiffies();
  int i = static_cast<int>(now/CLICK_HZ);
  int j = static_cast<int>(_now/CLICK_HZ);
  if ((i-j)>1) {
    // In next second
    double scale = 1000000;
    double xn1 = static_cast<double>(_x)/scale;
    double xn0 = static_cast<double>(_delta)/static_cast<double>(CLICK_HZ);
    double x = _beta * xn1 + (1-_beta) * xn0;
    _x = static_cast<int>(x*scale);
    _now = (j+1)*CLICK_HZ;  // Increase with 1 sec
    _delta = 1;
    
    // Update interface table
    assert(_if_no < _antTable->_interfaces.size());
    _antTable->_interfaces[_if_no].rate = _x * CLICK_HZ;
  } else {
    _delta = _delta + 1;
  }

  return(p);
}

enum { H_RATE, H_STATUS };
       
String
AntTrafficMonitor::read_handler(Element *e, void *thunk)
{
  AntTrafficMonitor *me = (AntTrafficMonitor *)e;
  String status = "";
  switch ((intptr_t)thunk) {
   case H_STATUS:
#ifdef HAVE_CLICK_KERNEL
      struct net_device *dev;
      for (dev=dev_base; dev; dev = dev->next) {
        struct in_ifaddr *ifa;
        if (dev->ip_ptr != NULL) {
          for (ifa = ((struct in_device*)(dev->ip_ptr))->ifa_list;ifa; ifa = ifa->ifa_next) {
            status += IPAddress(ifa->ifa_local).unparse() + "  ";
          }
        }
        if (netif_carrier_ok(dev) & (dev->flags & IFF_UP)) {
          status += String(dev->name) + " - UP\n";
        } else {
          status += String(dev->name) + " - DOWN\n";
        }
      }
#endif
     return status;
   case H_RATE:
    me->_packet_rate.update_time(); // drop rate after zero period
    me->_byte_rate.update_time(); // drop rate after zero period
    return me->_packet_rate.unparse() + " packets pr second\n" + 
           me->_byte_rate.unparse()   + " bytes pr second\n" + 
           cp_unparse_real10(me->_x*CLICK_HZ, 6) + " packets pr second\n";
   default:
    return "<error>\n";
  }
}

void
AntTrafficMonitor::add_handlers()
{
  add_read_handler("rate", read_handler, (void *)H_RATE);
  add_read_handler("status", read_handler, (void *)H_STATUS);
}


CLICK_ENDDECLS
ELEMENT_REQUIRES(AntTable)
EXPORT_ELEMENT(AntTrafficMonitor)
