/*
 * forwardant.{cc,hh} -- element looks up next-hop address in
 * Ant routing table.
 * Packet must be painted (See Paint()) with interface number (0..n)
 * 
 * Author: Ingebrigt Fuglem
 *
*/

#include <click/config.h>
#include "forwardant.hh"
#include <clicknet/ip.h>
#include <click/confparse.hh>
#include <click/error.hh>
#include <click/glue.hh>
#include <click/ipaddress.hh>
#include <click/vector.hh>
#include "anttable.hh"
#include <clicknet/udp.h>
#include <click/packet_anno.hh>
#include <click/straccum.hh>
#include <math.h>

CLICK_DECLS


ForwardAnt::ForwardAnt() : Element(1, 1), _antTable(0)
{

}

ForwardAnt::~ForwardAnt()
{
}

int
ForwardAnt::configure(Vector<String> &conf, ErrorHandler *errh)
{
	int res = cp_va_parse(conf, this, errh,
			cpElement, "AntTable element", &_antTable,
			cpEnd);
//	int ifs = _antTable->getNumberOfInterfaces();
//	click_chatter("ForwardAnt configured with %d interfaces", ifs);
//	set_noutputs(ifs);
  _id = 0;
	return res;
}

void
ForwardAnt::notify_noutputs(int n)
{
  set_noutputs(n < 2 ? 1 : n);
}

int
ForwardAnt::initialize(ErrorHandler *errh)
{
  if(_antTable == 0){
    return errh->error("no AntTable argument");
  }

  return 0;
}

Packet * update_checksum(WritablePacket *p) 
{
  WritablePacket *wp = 0;
  if (!wp) {
    if (!(wp = p->uniqueify()))
      return 0;
    else {
      click_ip *iph = wp->ip_header();
      click_udp *udph = wp->udp_header();
      
      uint16_t len = ntohs(udph->uh_ulen);
      int hlen = p->network_header_length();
     
      iph->ip_sum = 0;
      iph->ip_sum = click_in_cksum(p->network_header(), hlen);
      
      udph->uh_sum = 0;
      unsigned csum = click_in_cksum((unsigned char *)udph, len);
      udph->uh_sum = click_in_cksum_pseudohdr(csum, iph, len);
    }
  }
  
  return (wp ? wp : p);
}

Packet * add_cost(Packet *p, int cost) {
  WritablePacket *wp = 0;
  if (!(wp = p->uniqueify()))
    return 0;
  // Add cost to end of packet
  String sCost = "/" + String(cost);
  int udp_end = wp->length();
  wp = p->put(sCost.length());
  if (!wp)
    return 0;
  memcpy(wp->data() + udp_end, sCost.c_str(), sCost.length());
  
  // Update IP and UDP length
  click_ip *iph = wp->ip_header();
  click_udp *udph = wp->udp_header();
  uint16_t len = ntohs(udph->uh_ulen) + sCost.length();
  udph->uh_ulen = htons(len);
  iph->ip_len = htons(p->length());

  return update_checksum(wp);
}

Packet * ForwardAnt::send_notify(Packet *p, struct in_addr interface, int _flag) {
  // Interface status changed - notify ant sender
  WritablePacket *q = p->clone()->uniqueify();

  click_ip *ip = q->ip_header();
  click_udp *udp = q->udp_header();
  
  // Set src and dst ip-addr and port no
  ip->ip_src = interface;
  ip->ip_dst = p->ip_header()->ip_src;
  ip->ip_id = random();
  ip->ip_ttl = 9;
  udp->uh_sport = p->udp_header()->uh_dport;
  udp->uh_dport = p->udp_header()->uh_sport;
  
  // Use route record to create source routing
  uint8_t *oa = q->network_header();
  int hlen = q->network_header_length();

  StringAccum sourceroute;
  sourceroute << "SourceRoute: ";
	
  // Walk through options header
  for (int oi = sizeof(click_ip); oi < hlen; ) {
    unsigned type = oa[oi];
    if (type == IPOPT_EOL) {
      /* end of option list */
      break;
    } else if (type == IPOPT_RR) {
      // Route Record
      oa[oi] = IPOPT_LSRR;
      int last_ptr = oa[oi+2]-5; // points to last route record
      int first_ptr = 3;
      click_chatter("First RR[%d], Last RR[%d]", first_ptr, last_ptr);
      if (last_ptr >= 3) {
        // 1 or more
        struct in_addr dest;
        memcpy(&dest, oa + oi + last_ptr, 4);   // Use last as destination
        ip->ip_dst = dest;
        memcpy(oa + oi + last_ptr, &p->ip_header()->ip_src, 4);   // Put source in last position
        last_ptr -= 4;
      }
      while (last_ptr > first_ptr) {
        // swap 
        struct in_addr temp;
        memcpy(&temp, oa + oi + last_ptr, 4);
        memcpy(oa + oi + last_ptr, oa + oi + first_ptr, 4);
        memcpy(oa + oi + first_ptr, &temp, 4);
        sourceroute << IPAddress(temp).s().c_str() << " -> ";
        first_ptr += 4;
        last_ptr -= 4;
      }
      oa[oi+2] = 4;
      break;
    }
  }
  click_chatter("%s", sourceroute.cc());
  
  int udph = q->network_header_length() + sizeof(struct click_udp);   // Point to start of UDP payload
  String flag = String(_flag);
  memcpy(q->data() + udph, flag.c_str(), flag.length());

  q->set_dst_ip_anno(IPAddress(ip->ip_dst));
  q->set_ip_header(ip, sizeof(click_ip));
  q->set_network_header_length(p->network_header_length());
  
  click_chatter("Send NOTIFY for interface %s. Status: %s", IPAddress(interface).s().c_str(), flag.cc());
  return update_checksum(q);
}

bool in_route_record(Packet *p, struct in_addr interface) {
  const uint8_t *oa = p->network_header();
  int hlen = p->network_header_length();

//  click_chatter("Check if interface %s exsist in route record", IPAddress(interface).s().c_str());
  // Walk through options header
  for (int oi = sizeof(click_ip); oi < hlen; ) {
    unsigned type = oa[oi];
    int xlen = oa[oi + 1];  // length
    if (type == IPOPT_EOL) {
      /* end of option list */
      break;
    } else if (type == IPOPT_RR) {
      // Route Record
      int po = 4; // pointer
      while (po + 4 <= xlen) {
        if (memcmp(oa + oi + po-1, &interface.s_addr, 4) == 0)
          return true;
        po += 4;
      }
      return false;
    }
  }
  return false;
}

/* Check if interface is up */
bool is_if_up(struct in_addr interface) {
#ifdef HAVE_CLICK_KERNEL
  struct net_device *dev;
  for (dev=dev_base; dev; dev = dev->next) {
    struct in_ifaddr *ifa;
    if (dev->ip_ptr != NULL) {
      for (ifa = ((struct in_device*)(dev->ip_ptr))->ifa_list;ifa; ifa = ifa->ifa_next) {
        if (ifa->ifa_local == interface.s_addr) {
          return (netif_carrier_ok(dev) & (dev->flags & IFF_UP));
        }
      }
    }
  }
#else
  return true;
#endif
}

// Linksys stuff  
void get_linksys_status(int status[]) {
  char buffer[1024];
  FILE* fd = popen("robocfg show", "r");
  if (fd == NULL) {
    click_chatter("Unable to run robocfg");
  }
  
  int port = -1;
  while (NULL != fgets(buffer, sizeof(buffer), fd)) {
    // Port 0(W):  DOWN enabled stp: none vlan: 6 mac: 00:00:00:00:00:00
    char port_status[10];
    if (sscanf(buffer, "%*s%d%*s%s", &port, port_status) == 2) {
      if (port<6 && strcmp(port_status, "DOWN") == 0) {
	status[port] = 0;
      } else {
	status[port] = 1;
      }
    }
  }
  pclose (fd);
}

int isExplore(Packet *p) {
  int oi = p->network_header_length() + sizeof(struct click_udp);   // Point to start of UDP payload
  uint16_t len = p->length() - p->network_header_length();  // entire ip packet - ip header
  char payload[len];
  strncpy(payload, (char *)p->data() + oi, len);
  const char *pch = strtok(payload,"/");
  if (pch != NULL) {
    int explore = 0;
    cp_integer(String(pch), &explore);
    return explore;
  }
  return 0;
}

void
ForwardAnt::push(int, Packet *p) {
	
	
	
	
  		
  		
	int cost=0;
  assert(_antTable);
  
  int interface = 0;  // Will point to the choosen interface in entries list
  
  const IPAddress src_ip = IPAddress(p->ip_header()->ip_src);
  const IPAddress dest_ip = IPAddress(p->ip_header()->ip_dst);
  
  int explore = isExplore(p);
  click_chatter("FORWARD ant #%d: %s => %s", ntohs(p->ip_header()->ip_id), src_ip.s().c_str(), dest_ip.s().c_str());

  // Check status on all interfaces (Linksys)
  int if_status[6] = {0,0,0,0,0,0};
  get_linksys_status(if_status);

  if (_antTable->is_host(dest_ip)) {
    // Destination is this host. Send packet to last port.
    interface = _antTable->_interfaces.size();
    click_chatter("Send to local host. Output %d", interface);
  } else {
    struct in_addr in_if = _antTable->getInterface(PAINT_ANNO(p)).address.in_addr();
    click_chatter("on incomming interface %s (%d)", 
		  IPAddress(in_if).s().c_str(), PAINT_ANNO(p));
    int in_if_ptr = -1;
    
    int ptr = _antTable->get_first(src_ip, dest_ip);

    // Create a vector with all candidate interfaces
    // Not include interfaces that are used before
    Vector<int> candidates;
    while (ptr >= 0) {
      AntTable::ATEntry *at_entry = &_antTable->_ant_routes[ptr];
      // Add interface if it has not been used before, except the incomming interface
      if (in_if == at_entry->interface.in_addr())
        in_if_ptr = ptr;
      else if (!in_route_record(p, at_entry->interface.in_addr()))
      {
	//      	click_chatter("-- 2.attempt --");
	//      	in_route_record(p, at_entry->interface.in_addr());
      	
        //bool if_up = is_linksys_if_up(at_entry->interface.in_addr());
        bool if_up = (bool)if_status[at_entry->physical_port];
        click_chatter("Interface %s is %s", at_entry->interface.s().c_str(),(if_up?"UP":"DOWN"));
        if (if_up) {
          candidates.push_back(ptr);
        }
        if (if_up != at_entry->if_carrier_ok) {
          int n = noutputs();
          // Send a new packet back with information about link status changed
          if (n > _antTable->getNumberOfInterfaces()) {
            Packet *q = send_notify(p, at_entry->interface.in_addr(), 3+(int)if_up);
            checked_output_push(n-1, q);
          }
          at_entry->if_carrier_ok = if_up;
        }
      }
      ptr = _antTable->get_next(ptr);
    }
    // If all interfaces has been tried - go back the way you came
    if (candidates.empty() && in_if_ptr >= 0)
      candidates.push_back(in_if_ptr);

    assert(!candidates.empty());
    
    // Check TTL
    if (p->ip_header()->ip_ttl <= 1) {
      Packet *q = send_notify(p, in_if, 2);
      checked_output_push(noutputs()-1, q);
    }

#ifdef CLICK_USERLEVEL
    for (int i = 0; i<candidates.size();i++)
      click_chatter("candidates = %s (%d)",  
		    _antTable->_ant_routes[candidates[i]].interface.s().c_str(), 
		    candidates[i]);
#endif

    // Do the selection
    
    float sum_phero = 0;
    if (explore != 0) {
      // Explorer ant - all pigs are equal
      ptr = random()%candidates.size();
      click_chatter("Explorer ant: ptr = %d", ptr);      
      interface = _antTable->_ant_routes[candidates[ptr]].port;
    } else {
      for (int i=0; i < candidates.size(); i++) {
        sum_phero += _antTable->_ant_routes[candidates[i]].pheromone;
        click_chatter("IF[%d]: pher = %f", _antTable->_ant_routes[candidates[i]].port, _antTable->_ant_routes[candidates[i]].pheromone);
      }
    
      if (sum_phero == 0){
        // No pheromones updated yet - choose random
        ptr = random()%candidates.size();
        click_chatter("uninit: ptr = %d", ptr);      
        interface = _antTable->_ant_routes[candidates[ptr]].port;
      } else {

        ptr = 0;
        double samp = static_cast<double>(sum_phero * random())/static_cast<double>(RAND_MAX);
        double tmp = _antTable->_ant_routes[candidates[ptr]].pheromone;
        while (tmp < samp && ptr < (candidates.size()-1)) {
          ptr++;
          tmp += _antTable->_ant_routes[candidates[ptr]].pheromone;
        }
        click_chatter("selected: ptr = %d", ptr);      
        interface = _antTable->_ant_routes[candidates[ptr]].port;
      }
    }
    cost = _antTable->_interfaces[interface].cost;
    click_chatter("*** selected interface %s (%d) with cost %d ***", 
		  _antTable->_ant_routes[candidates[ptr]].interface.s().c_str(), 
		  interface, cost);
   
  }		
  		p = add_cost(p, cost);
  		
  		
  
  		
  output(interface).push(p);
}


CLICK_ENDDECLS

ELEMENT_REQUIRES(AntTable)
EXPORT_ELEMENT(ForwardAnt)
