/*
 * SourceRoute.{cc,hh} -- element processes IP Timestamp and Record Route
 * options
 * Robert Morris, Eddie Kohler
 *
 * Copyright (c) 1999-2000 Massachusetts Institute of Technology
 * Copyright (c) 2003 International Computer Science Institute
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, subject to the conditions
 * listed in the Click LICENSE file. These conditions include: you must
 * preserve this copyright notice, and you cannot mention the copyright
 * holders in advertising related to the Software without their permission.
 * The Software is provided WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED. This
 * notice is a summary of the Click LICENSE file; the license in that file is
 * legally binding.
 */

#include <click/config.h>
#include <click/confparse.hh>
#include <click/error.hh>
#include <click/glue.hh>
#include "sourceroute.hh"
#include <clicknet/ip.h>
#include <click/packet_anno.hh>
CLICK_DECLS

SourceRoute::SourceRoute()
{
  _drops = 0;
	//MOD_INC_USE_COUNT;
	add_input();
	add_output();
}

SourceRoute::~SourceRoute()
{
	//MOD_DEC_USE_COUNT;
}

int
SourceRoute::configure(Vector<String> &conf, ErrorHandler *errh)
{
  if (cp_va_parse(conf, this, errh,
		  cpIPAddress, "local IP address", &_my_addr,
		  cpEnd) < 0)
    return -1;
  return 0;
}

void
SourceRoute::notify_noutputs(int n)
{
  // allow 2 outputs -- then packet is pushed onto 2d output instead of
  // dropped
  set_noutputs(n < 2 ? 1 : 2);
}

Packet *
SourceRoute::handle_options(Packet *p)
{
  WritablePacket *wp = 0;
  const uint8_t *oa = p->network_header();
  int hlen = p->network_header_length();
  
  int oi;
  for (oi = sizeof(click_ip); oi < hlen; ) {
    // handle one-byte options
    unsigned type = oa[oi];
    if (type == IPOPT_NOP) {
      oi++;
      continue;
    } else if (type == IPOPT_EOL)
      /* end of option list */
      break;
    
    // otherwise, get option length
    int xlen = oa[oi + 1];
    if (xlen < 2 || oi + xlen > hlen) {
      // bad length
      oi++;       // to point at length
      goto send_error;
    } else if (type != IPOPT_LSRR && type != IPOPT_SSRR) {
      // not for us to process
      oi += xlen;
      continue;
    }
    
    
    // need a writable packet
    if (!wp) {
      if (!(wp = p->uniqueify()))
	return 0;
      oa = wp->network_header(); // may have changed due to packet copy
    }
    uint8_t *woa = wp->network_header();
    
    // p will point to the next router
    int p = oa[oi+2] - 1;
    if (p >= 3 && p + 4 <= xlen) {
      struct in_addr next_hop;
      
      // Get next router from options
      memcpy(&next_hop, oa + oi + p, 4);
      
      //	IPAddress dst = wp->dst_ip_anno();
      click_ip *iph = wp->ip_header();
      IPAddress dst = wp->ip_header()->ip_dst;
      //click_chatter("Dest = %s", dst.s().c_str());

      if (dst.in_addr() == _my_addr || type == IPOPT_LSRR) {
	// Update destination and move pointer
	click_chatter("Source routing. Changed destination from %s to %s",
		      dst.s().c_str(), IPAddress(next_hop).s().c_str());
	// Record route
	memcpy(woa + oi + p, dst.data(), 4);
	// Move pointer
	woa[oi+2] += 4;
	// Change destination address
	wp->set_dst_ip_anno(next_hop);
	iph->ip_dst = next_hop;
      } else {
        click_chatter("Destination address %s is not me, so packet is discarded because of strict routing", 
        dst.s().cc());
	return 0;
      }
			
      oi += xlen;
    } else if (p != xlen) {
      oi += 2;
      goto send_error;
    } else if (p == xlen) {
      // We are at the end of the list, I am the final destination
    }
  }

  if (wp) {
    click_ip *iph = wp->ip_header();
    int hlen = iph->ip_hl << 2;
    iph->ip_sum = 0;
    iph->ip_sum = click_in_cksum((unsigned char *)iph, hlen);
  }
  
  return (wp ? wp : p);
  
 send_error:
  _drops++;
  SET_ICMP_PARAMPROB_ANNO(p, oi);
  checked_output_push(1, p);
  return 0;
}

Packet *
SourceRoute::simple_action(Packet *p)
{
  const click_ip *ip = p->ip_header();
  assert(ip);

  unsigned hlen = ip->ip_hl << 2;
  //  click_chatter("SourceRoute handles packet. Destination address is %x", ip->ip_dst.s_addr);
  if (hlen <= sizeof(click_ip) || (p = handle_options(p)))
    return(p);
  return(0);
}

static String
SourceRoute_read_drops(Element *xf, void *)
{
	SourceRoute *f = (SourceRoute *)xf;
	return String(f->drops()) + "\n";
}

void
SourceRoute::add_handlers()
{
	add_read_handler("drops", SourceRoute_read_drops, 0);
}

CLICK_ENDDECLS
EXPORT_ELEMENT(SourceRoute)
ELEMENT_MT_SAFE(SourceRoute)
